#!/usr/bin/env python3

from sys import argv, stdin, version_info, version, executable
import json
import jedi
import inspect
import pydoc
import re
import keyword
import os

cr_version = int(os.getenv("CR_VERSION", "0"))

def is_identifier_char(s):
	if s == "": return False
	if hasattr(str, "isidentifier"): return ("_"+s).isidentifier()
	if s.isalnum(): return True
	s = s.replace("_", "")
	return s == "" or s.isalnum()

def standard_type(word):
	map = {
		"int": "int",
		"integer": "int",
		"str": "str",
		"string": "str",
		"dict": "dict",
		"dictionary": "dict",
		"number": "number",
		"bool": "bool",
		"boolean": "bool",
		"character": "str",
		"list": "list",
		"tuple": "tuple",
		"pair": "tuple",
		"set": "set",
		"true": "bool",
		"false": "bool",
		"-1": "int",
		"index": "int",
		"none": "None",
		"iterator": "iterator",
	}
	if word in map:
		return map[word]
	else:
		return None

def process_full_name(completion, item):
	full_name = ""
	name = completion.name
	try:
		full_name = completion.full_name
	except AttributeError:
		full_name = name
	if not full_name: full_name = name
	dot = full_name.rfind(".")
	if dot == -1:
		try:
			class_name = completion._name._obj.name
			if class_name:
				full_name = class_name + "." + full_name
		except:
			if full_name != name:
				dot = len(full_name)
				full_name = full_name + "." + name
	if dot != -1 and full_name[dot+1:] != name:
		t = standard_type(full_name[dot+1:])
		if t:
			item["type"] = t
		full_name = full_name[:dot+1] + name
		if item["kind"] == "class":
			item["kind"] = ""
	item["full_name"] = full_name

def insertion_for_params(params, add_placeholders_for_none_params = False):
	if params == None:
		if add_placeholders_for_none_params:
			return "(<##>)"
		else:
			return "()"
	insertion = ""
	separator = False
	optional = 0
	for name in params:
		if len(name) == 0:
			continue
		name = name.replace(" = ", "=")
		bracket = False
		if name.startswith("["):
			bracket = True
			name = name[1:]
			if name.endswith("]"):
				name = name[:-1]
			name = name.strip()
		search = name.find("=")
		ellipsis = False
		if optional or search != -1 or bracket:
			if separator:
				insertion += " [, "
			else:
				insertion += "[ "
			optional += 1
		elif name == "...":
			ellipsis = True
			if separator:
				insertion += ", "
		elif separator:
			insertion += "#>, "
		if optional == 0 and separator and ellipsis == False:
			insertion += "<#"
		if separator == False:
			separator = True
		insertion += name
	if optional > 0:
		insertion += " "
	for x in range(0, optional):
		insertion += "]"
	if separator:
		insertion = "<#"+insertion+"#>"
	if insertion == "<#...#>":
		insertion = "<##>"
	insertion = "(" + insertion + ")"
	return insertion

def process_insertion(completion, item):
	if item["kind"] not in ["function", "method", "class"]:
		return
	tree_name = None
	try: tree_name = completion._name.tree_name
	except: pass
	alternate_signature_params = []
	params = get_params(item["full_name"], item, tree_name, completion, alternate_signature_params)
	item["insertion"] = insertion_for_params(params, item["kind"] == "class")
	if len(alternate_signature_params) != 0:
		item["alternate_insertions"] = []
		for alternate_params in alternate_signature_params:
			insertion = insertion_for_params(alternate_params, item["kind"] == "class")
			if insertion not in item["alternate_insertions"]:
				item["alternate_insertions"].append(insertion)
def get_params(full_name, item = None, tree_name = None, completion = None, alternate_signature_params = []):
	params = None
	try:
		obj = pydoc.locate(full_name, 0)
		if obj and callable(obj):
			signature = inspect.signature(obj)
			params = []
			for p in signature._parameters.values():
				param = str(p)
				pos = param.find("=<")
				if pos != -1:
					param = "[" + param[:pos] + "]"
				params.append(param)
			if len(params) and params[0] == "self":
				if item != None and item["kind"] == "function" and item["full_name"].find(".") != -1:
					item["kind"] = "method"
				del params[0]
	except:
		params = None
	if params == None:
		try:
			for p in tree_name.get_definition().get_params():
				text = p.name.value
				if params == None:
					params = []
					if text == "self":
						continue
				if text.startswith("__") and len(text) > 2 and not text.endswith("__") and full_name.startswith("builtins."):
					text = text[2:]
				code = p.get_code(include_prefix=False, include_comma=False)
				optional = code.find(": Optional[")
				if optional != -1:
					text = "[" + text[:optional] + "]"
				params.append(text)
		except:
			pass
	if params == None and completion != None:
		doc = cached_completion_docstring(completion)
		if doc:
			for match in re.finditer(r'(?m)^' + re.escape(item["string"]) + r'\((.*)\)', doc):
				params_code = match.group(1)
				if params_code.find(":") != -1:
					continue
				params_split = [p.strip() for p in params_code.split(",")]
				if params == None:
					params = params_split
				elif params != params_split:
					alternate_signature_params.append(params_split)
	return params

def docstring2html(docstring):
	while True:
		end = docstring.find("\n")
		if end == -1: break
		line = docstring[:end]
		if line.find(" -> ") != -1 or line.find(" --> ") != -1 or len(line) == 0 or line[-1] == ')':
			docstring = docstring[end+1:]
		else:
			break
	docstring = docstring.replace('<', '&lt;')
	docstring = docstring.replace('>', '&gt;')
	docstring = re.sub('``(.*?)(``|\'\')', '<code>\\1</code>', docstring)
	docstring = re.sub('`(.+?)(`|\')', '<code>\\1</code>', docstring)
	docstring = re.sub('\n:type\\s+\\w+\\s*:\\s*(.*?)(?=\n|$)', ' Type: <code>\\1</code>', docstring)
	docstring = re.sub('\n:param\\s+(\\w+)\\s*:', '<br><b><code>\\1</code></b>:', docstring)
	docstring = re.sub(':rtype:', 'Return type: ', docstring)
	docstring = re.sub(':\\w+:', '', docstring)
	docstring = re.sub(r'^\s*[=\-~*][=\-~*\s]*$', '<br>', docstring, flags=re.MULTILINE)
	docstring = docstring.replace("\n\n", "<br>")
	return docstring

last_cached_docstring_completion = None
last_cached_docstring = ""
def cached_completion_docstring(completion):
	try:
		global last_cached_docstring_completion
		global last_cached_docstring
		if completion == last_cached_docstring_completion:
			return last_cached_docstring
		last_cached_docstring = completion.docstring(raw=False)
		last_cached_docstring_completion = completion
		return last_cached_docstring
	except:
		return ""

def process_description(completion, item):
	if item["kind"] == "param":
		try:
			full_name = ".".join(completion._name._wrapped_param_name.function_value.get_qualified_names())
			tree_name = completion._name._wrapped_param_name.function_value.name.tree_name
			alternate_signature_params = []
			params = get_params(full_name, None, tree_name, None, alternate_signature_params)
			if len(alternate_signature_params) == 0:
				insertion = insertion_for_params(params).replace("#>", "").replace("<#", "")
			else:
				insertion = "(...)"
			function_name = completion._name._wrapped_param_name.function_value.tree_node.name.value
			item["plain_description"] = function_name + insertion
			item["description"] = "<code>" + item["plain_description"] + "</code>"
		except:
			pass
		return
	elif item["kind"] == "keyword":
		return
	s = ""
	if item["kind"] != "module":
		try:
			object = pydoc.locate(item["full_name"], 0)
			if inspect.isroutine(object):
				s = pydoc.getdoc(object)
				if inspect.ismethoddescriptor(object):
					item["kind"] = "method"
				lines = s.splitlines()
				i = 0
				while i < len(lines):
					lines[i] = lines[i].strip()
					if len(lines[i]) and lines[i][0] == '|':
						del lines[i]
					else:
						i += 1
				s = "\n".join(lines)
		except:
			pass
	if not s:
		s = cached_completion_docstring(completion)
		if s and not s.startswith(item["full_name"]) and not s.startswith(item["string"]):
			s = ""
	parent_name = ""
	dot = item["full_name"].rfind(".")
	if dot != -1:
		prev_dot = item["full_name"].rfind(".", 0, dot)
		if prev_dot != -1:
			parent_name = item["full_name"][prev_dot+1:dot]
		else:
			parent_name = item["full_name"][:dot]
	if item["full_name"] == "sys.version":
		if s: s += "\n\n"
		s += version + "\n\n" + executable
	item["plain_description"] = s
	s = docstring2html(s)
	if parent_name:
		if cr_version > 50000:
			info = "<b><code>" + parent_name + "</code></b>"
		else:
			info = "<code>" + parent_name + "</code>"
		if s: s = info + " &ndash; " + s
		item["parent_name"] = parent_name
	if s: item["description"] = s

def process_kind(completion, item):
	if item["kind"] == "class":
		try:
			obj = pydoc.locate(item["full_name"], 0)
			if issubclass(obj, BaseException):
				item["kind"] = "exception"
		except: pass
	elif item["kind"] == "statement":
		if item["string"].isupper():
			item["kind"] = "constant"
		else:
			try:
				obj = pydoc.locate(item["full_name"], 0)
				if inspect.ismethod(obj) or inspect.ismethoddescriptor(obj):
					item["kind"] = "method"
					return
				elif inspect.isfunction(obj) or callable(obj):
					item["kind"] = "function"
					return
			except: pass
			item["kind"] = "variable"

def guess_type(completion, item):
	if "plain_description" in item:
		types = set()
		description = item["plain_description"]
		for match in re.finditer(r'(?i)\b(return(s|ing)? )(a |an |the )?(\w+)', description):
			word = match.group(4).lower()
			t = standard_type(word)
			if t and t != "None":
				types.add(t)
			else:
				types.clear()
				break
		if len(types) == 1:
			for x in types: return x
	return None

def process_type(completion, item):
	if (item["kind"] == "function" or item["kind"] == "method") and not item["string"].startswith("__"):
		map = {
			"builtins.len": "int",
			"builtins.abs": "number",
			"builtins.bin": "str",
			"builtins.callable": "bool",
			"builtins.chr": "str",
			"builtins.delattr": "None",
			"builtins.format": "str",
			"builtins.hasattr": "bool",
			"builtins.hash": "int",
			"builtins.hex": "str",
			"builtins.id": "int",
			"builtins.input": "str",
			"builtins.isinstance": "bool",
			"builtins.issubclass": "bool",
			"builtins.oct": "str",
			"builtins.ord": "int",
			"builtins.pow": "number",
			"builtins.print": "None",
			"builtins.repr": "str",
			"builtins.round": "number",
			"builtins.setattr": "None",
			"builtins.sorted": "list",
			"builtins.sum": "number",
			"builtins.ascii": "str",
			"builtins.dir": "list",
			"builtins.iter": "iterator",
			"builtins.vars": "dict",
			"list.append": "None",
			"float.as_integer_ratio": "tuple",
			"int.bit_length": "int",
			"str.capitalize": "str",
			"str.casefold": "str",
			"str.center": "str",
			"str.removeprefix": "str",
			"str.removesuffix": "str",
			"list.clear": "None",
			"int.conjugate": "int",
			"str.encode": "str",
			"int.as_integer_ratio": "tuple",
			"float.conjugate": "float",
			"str.expandtabs": "str",
			"list.extend": "None",
			"float.fromhex": "float",
			"dict.fromkeys": "dict",
			"float.hex": "str",
			"list.insert": "None",
			"str.join": "str",
			"str.ljust": "str",
			"str.lower": "str",
			"str.lstrip": "str",
			"str.maketrans": "dict",
			"str.partition": "tuple",
			"dict.popitem": "tuple",
			"list.remove": "None",
			"str.replace": "str",
			"list.reverse": "None",
			"str.rjust": "str",
			"str.rpartition": "tuple",
			"str.rstrip": "str",
			"str.strip": "str",
			"str.swapcase": "str",
			"str.title": "str",
			"int.to_bytes": "bytes",
			"str.translate": "str",
			"str.upper": "str",
			"str.zfill": "str",
			"list.copy": "list",
			"tuple.index": "int",
			"set.add": "None",
			"bool.as_integer_ratio": "tuple",
			"bool.bit_length": "str",
			"set.clear": "None",
			"bool.conjugate": "int",
			"set.copy": "set",
			"set.difference": "set",
			"set.difference_update": "None",
			"set.discard": "None",
			"set.intersection": "set",
			"set.intersection_update": "None",
			"set.issubset": "bool",
			"set.issuperset": "bool",
			"set.remove": "None",
			"set.symmetric_difference": "set",
			"set.symmetric_difference_update": "None",
			"bool.to_bytes": "bytes",
			"set.union": "set",
			"set.update": "None",
			"list.index": "int",
			"str.find": "int",
			"str.format": "str",
			"str.format_map": "str",
			"str.rfind": "int",
			"str.index": "int",
			"str.rindex": "int",
			"dict.copy": "dict",
		}
		if item["full_name"] in map:
			item["type"] = map[item["full_name"]]
		elif item["full_name"].startswith("builtins.") and item["full_name"][9:] in map:
			item["type"] = map[item["full_name"][9:]]
		if "type" in item:
			return
		typereg = r'(?m)\) -?-> ([a-z]+(\[[\w, ]+\])?|[A-Za-z]+)( or None)?($|[\.,]($|\s))'
		if "plain_description" in item:
			match = re.search(typereg, item["plain_description"])
			if match:
				item["type"] = match.group(1)
		if not "type" in item:
			doc = cached_completion_docstring(completion)
			if doc:
				match = re.search(typereg, doc)
				if match:
					item["type"] = match.group(1)
		if "type" in item:
			t = item["type"]
			open_bracket = t.find("[")
			close_bracket = t.find("]")
			if open_bracket != -1 and close_bracket != -1 and open_bracket < close_bracket:
				first_word = t[:open_bracket].lower()
				if first_word == "optional":
					t = t[open_bracket+1:close_bracket]
				else:
					t = first_word
			t = standard_type(t)
			if t != None:
				item["type"] = t
			else:
				del item["type"]
		if not "type" in item:
			t = guess_type(completion, item)
			if t: item["type"] = t
def get_position(source, line, column):
	source_lines = source.splitlines(True)
	if len(source_lines) == 0 or (len(source_lines[-1]) and source_lines[-1][-1] == "\n"):
		source_lines.append("")
	pos = 0
	valid = False
	for idx in range(line):
		if idx+1 == line:
			valid = len(source_lines[line-1].rstrip("\r\n")) >= column-1
			pos += column
			break
		if idx < len(source_lines):
			pos += len(source_lines[idx])
	if valid:
		return pos
	return -1

def make_class_name_item(item):
	class_item = item.copy()
	if "insertion" in class_item: del class_item["insertion"]
	if "parent_name" in item and item["parent_name"] != item["string"]:
		class_item["plain_description"] = item["parent_name"]
		class_item["description"] = "<b><code>" + item["parent_name"] + "</code></b>"
	else:
		if "plain_description" in class_item: del class_item["plain_description"]
		if "description" in class_item: del class_item["description"]
	return class_item

def process_completion(completion):
	name = completion.name
	item = {
		"string": name,
		"showImmediately": showImmediately,
		"docsUUID": "FF667E76-BF73-4C18-AD70-EABAC0861D92",
		"priority": 0 if name.startswith("__") else 1
	}
	kind = completion.type
	if isinstance(kind, str):
		item["kind"] = kind
	try: process_full_name(completion, item)
	except: pass
	try: process_kind(completion, item)
	except: pass
	try: process_insertion(completion, item)
	except: pass
	try: process_description(completion, item)
	except: pass
	try: process_type(completion, item)
	except: pass
	return item

filename = argv[1]
source = stdin.read()
if len(argv) > 2:
	line = int(argv[2])
	column = int(argv[3])-1
else:
	lines = source.splitlines()
	line = len(lines)
	column = len(lines[-1])
pos = get_position(source, line, column)
while pos > 0 and is_identifier_char(source[pos-1]):
	pos -= 1
	column -= 1
while pos < len(source) and is_identifier_char(source[pos]):
	source = source[:pos] + source[pos+1:]
script = jedi.Script(code=source, path=filename)
completions = script.complete(line=line, column=column)
if len(completions) == 0:
	# Jedi returns no completions if it doesn't know the object's type, which is unhelpful.
	# If completing after a dot, make Jedi return the set of completions for common builtin types.
	if pos != -1:
		while pos > 0 and source[pos-1].isspace(): pos -= 1
		if pos > 0 and source[pos-1] == '.':
			generic = "[list(),dict(),str(),int(),float(),tuple(),set()][_x_]."
			completing_chars = 0
			linestart = source.rfind("\n", 0, pos-completing_chars)+1
			match = re.match(r'\s*((\w+)\s+)$', source[linestart:pos-completing_chars])
			if match and keyword.iskeyword(match.group(2)):
				completing_chars += len(match.group(1))
			source = source[:pos-completing_chars] + generic + source[pos:]
			column += len(generic) - completing_chars
			script = jedi.Script(code=source, path=filename)
			completions = script.complete(line=line, column=column)
output = []
showImmediately = False
if len(argv) > 4:
	context = argv[4]
	if re.search(r'\.\s*$', context):
		if not re.search(r'\b\d\.\s*$', context):
			showImmediately = True
for completion in completions:
	try:
		item = process_completion(completion)
		output.append(item)
		if "alternate_insertions" in item:
			for alternate_insertion in item["alternate_insertions"]:
				alternate_item = item.copy()
				alternate_item["insertion"] = alternate_insertion
				output.append(alternate_item)
		try:
			if item["kind"] == "class":
				class_item = make_class_name_item(item)
				if class_item:
					output.append(class_item)
		except: pass
		extra_completions = completion._same_name_completions
		if len(extra_completions):
			if item["kind"] in ["function", "method"] and item["full_name"].find(".") != -1 and not item["string"].startswith("__"):
				for extra_completion in extra_completions:
					if extra_completion in completions: continue
					try:
						extra_item = process_completion(extra_completion)
						if extra_item != item:
							output.append(extra_item)
					except:
						pass
	except:
		pass
print(json.dumps(output))
